@@ -14,7 +14,7 @@ sub moos {
                               ------   ------
                      ^__^   /                 \   ^__^
              _______/(oo)  /                   \  (oo)\_______
-         /\/) ingy  /(__)                         (__)\  mst  )\/\
+         /\/( ingy  /(__)                         (__)\  mst  )\/\
             | w----||                              /  ||----w |
             ||     ||                                 ||     ||
 EOEOMOO
